//     ( )
// Proteus (lawnmower-man@mail.ru)

#include<stdio.h>
#include<stdlib.h>

//!  
template<class T>class List{
	T *first; //!<  
	T *last;  //!<  
public:
	List():first(0) {}
	~List() {while(first){T *n=first->next;delete first;first=n;}}
	//!   
	operator T*()  {return first;}
	//!    
	void operator +=(T *x) {if(first) last->next=x; else first=x; last=x,x->next=0;}
};

//!  
struct node_st{
	node_st *next;
	node_st *semi;	//!<  
	int id;			//!<  
	char sym;		//!<  
	bool mark;		//!<   
	node_st(char s):sym(s),id(-1),semi(0) {}
};

//!  
struct link_st{
	link_st *next;
	node_st *from;	//!<  
	node_st *to;	//!<  
	bool mark;		//!<    
	link_st(node_st *f,node_st *t):mark(false),from(f),to(t) {}
};

List<node_st>nodes;	//!<  
List<link_st>links;	//!<  

//!    
void AddNode(char sym)
{
	nodes+=new node_st(sym);
}

//!      
node_st *GetNode(char sym)
{
	for(node_st *n=nodes; n; n=n->next) if(n->sym==sym) return n;
	return 0;
}

//!    
void AddLink(char s1,char s2)
{
	links+=new link_st(GetNode(s1),GetNode(s2));
}

//!     
void UnsetNodesMark()
{
	for(node_st *n=nodes; n; n=n->next) n->mark=false;
}

//!      ,  
//! \param node  
//! \param index  
void TreeTraversal(node_st *node,int &index)
{
	node->mark=true;
	node->id=index++;
	for(link_st *l=links; l; l=l->next) if(l->from==node && !l->to->mark) {
		l->mark=true;
		TreeTraversal(l->to,index);
	}
}

//!      
void TreeTraversal()
{
	int index=1;
	UnsetNodesMark();
	TreeTraversal(nodes,index);
}

//!  ,  
//! \param it    
//! \param node      
void FindSemi(node_st *it,node_st *node)
{
	//      
	if(it->mark) return; it->mark=true;
	//     ,      ,   ,  
	if(it->id<node->id) { if(!node->semi || it->id<node->semi->id) node->semi=it; return;}
	//      
	for(link_st *l=links; l; l=l->next) if(l->to==it) FindSemi(l->from,node);
}


//!  
void FindSemi()
{
	for(node_st *n=nodes; n; n=n->next) {
		UnsetNodesMark();
		FindSemi(n,n);
	}
}


//!   ,   
//!     ,     (    )
//! \param from   
//! \param to      
//! \result true     
bool FindDomi(node_st *from,node_st *to)
{
	bool res=false;
	if(from==to) return true; //  ,    
	for(link_st *l=links; l; l=l->next) if(l->mark && l->from==from && FindDomi(l->to,to)) {
		//      
		res=true;
		if(l->to->semi->id<to->semi->id) to->semi=l->to->semi;
	}
	return res;
}

//!   
void FindDomi()
{
	for(node_st *n=nodes; n; n=n->next) FindDomi(n->semi,n);
}

//================================================================

//!    
void DumpGraph(const char *filename)
{
	FILE *file=fopen(filename,"wt");
	fprintf(file,"digraph{\n" "size=\"5,5\"\n");
	for(node_st *n=nodes; n; n=n->next) fprintf(file,"  %c [label=\"%c [%d,%c]\"];\n",n->sym,n->sym,n->id,n->semi?n->semi->sym:' ');
	for(link_st *l=links; l; l=l->next) fprintf(file,"  %c -> %c%s;\n",l->from->sym,l->to->sym,l->mark?"":" [style=dotted]");
	fprintf(file,"}\n");
	fclose(file);
}

//!     
void DumpDomi(const char *filename)
{
	FILE *file=fopen(filename,"wt");
	fprintf(file,"digraph{\n");
	for(node_st *n=nodes; n; n=n->next) if(n->semi) fprintf(file,"  %c -> %c;\n",n->semi->sym,n->sym);
	fprintf(file,"}\n");
	fclose(file);
}

//!    
void Output()
{
	for(node_st *n=nodes; n; n=n->next) if(n->semi) printf("d( %c )= %c\n",n->sym,n->semi->sym);
}


//================================================================

//!    
void MakeGraph()
{
	const char *nlist="RCFIKGJBEHADL";
	const char *llist="RC+CF+CG+FI+GI GJ+JI IK+KI KR RB+RA BE+BA+BD EH+AD+HE HK DL+LH ";
	while(*nlist) AddNode(*nlist++);
	while(*llist) AddLink(llist[0],llist[1]),llist+=3;
}

//!   
void MakeRandGraph()
{
	const int size=20;
	int n;
	for(n=0; n<size; n++) AddNode('A'+n);
	for(n=0; n<size*3; n++) AddLink('A'+rand()%size,'A'+rand()%size);
}



int main()
{
	if(1) MakeGraph(); else MakeRandGraph();	//   
	TreeTraversal();	// Step 1
	FindSemi();			// Step 2
	FindDomi();			// Step 3
	DumpGraph("graph.dot");
	DumpDomi ( "domi.dot");
	Output();
	return 0;
}


